/*
 * This code is an example taken directly from the Sgi driver manual.  It has
 * been modified to work with the Symmetricom BC635 Time and Frequency Processor
 * board to test receipt of user level interrupts.  All setup of the board and
 * enabling of interrupts is done by a separate menu driven program.  This just
 * tests receipt of interrupts.
 *
 *-------------------
 *
 * pci40_uli.c - PCI User Level Interrupt (ULI) test using the
 *               Greenspring PCI40 IP carrier card to generate
 *               interrupts.
 *
 * This version for Onyx 2/Origin 2000 systems (Origin 200 systems
 * will have a different hwgraph path.)
 *
 * link with -luli
 *
 * Make sure that the latest 6.5 REACT/pro, PCI and kernel
 * roll-up patches are installed.
 *
 */
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/prctl.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/syssgi.h>
#include <sys/sysmp.h>
#include <sched.h>
#include <sys/uli.h>

#define CYCLE_COUNTER 1

#define     INTRPATH   "/hw/module/001c02/Ibrick/xtalk/14/pci/1/ppb_16/0/intr"
#define     BC635_PATH    "/hw/module/001c02/Ibrick/xtalk/14/pci/1/ppb_16/0/mem.le"
#define     BC635_SIZE 0x40

#define PCI_INTA   0x01
#define NUM_INTS   1000
#define BAD_RESPONSE  30
#define PROC       0
extern int errno;
volatile int intr = 0;
static void *ULIid;

struct stfp_registers
{
  uint32_t timereq;			/* time request (for time[] regs) */
  uint32_t eventreq;			/* time request (for event[] regs) */
  uint32_t unlock;			/* time capture lockout release */
  uint32_t reserved;
  uint32_t control;			/* misc. control register */
  uint32_t ack;				/* acknowledge register */
  uint32_t mask;			/* interrupt mask register */
  uint32_t intstat;			/* interrupt status register */
  uint32_t minstrb;			/* minor strobe time */
  uint32_t majstrb;			/* major strobe time */
  uint32_t reserved2[2];
  uint32_t time[2];			/* minor/major time hold */
  uint32_t event[2];			/* minor/major event */
};

volatile struct stfp_registers*  registers;

uint32_t* interrupt_arg;

/* definitions for timer */
typedef unsigned long long iotimer_t;
__psunsigned_t phys_addr, raddr;
unsigned int cycleval;
volatile iotimer_t begin_time, end_time, *timer_addr;
int timer_fd, poffmask;
float usec_time;

float longest_response = 0.0;
float average_response = 0.0;


static void
intrfunc(void *arg)
{
  /* Set the global flag indicating to the main thread that an
   * interrupt has occurred, and wake it up
   */
  intr++;
  /*
   * clear the interrupt on the mothercard by clearing CNTRL0
   * adding 1 to offset for big endian access
   */
  registers->intstat = 0xffffffff;
}


main(int argc, char *argv[])
{
  int     fd;
  int     pci_fd;
  volatile uint_t x;
  struct timespec wait_time;
  int last_intr;

  interrupt_arg = (uint32_t*)malloc(128);
  if (interrupt_arg == (uint32_t*)NULL)
  {
    perror("malloc");
    exit(1);
  }

  if (mlockall( MCL_CURRENT | MCL_FUTURE ))
  {
    perror ("mlockall");
  }

#if CYCLE_COUNTER
  /*
   * memory map the hardware cycle-counter
   */
  poffmask = getpagesize() - 1;
  phys_addr = syssgi(SGI_QUERY_CYCLECNTR, &cycleval);
  raddr = phys_addr & ~poffmask;
  timer_fd = open("/dev/mmem", O_RDONLY);
  timer_addr = (volatile iotimer_t *)mmap(0, poffmask, PROT_READ,
                                          MAP_PRIVATE, timer_fd, (off_t)raddr);
  timer_addr = (iotimer_t *)((__psunsigned_t)timer_addr +
                             (phys_addr & poffmask));
#endif

  /*
   * open the PCI user interrupt device/vertex
   */
  fd = open(INTRPATH, O_RDWR);
  if (fd < 0 )
  {
    perror(INTRPATH);
    exit (1);
  }

  /*
   * open the memory space for device registers
   */
  pci_fd = open(BC635_PATH, O_RDWR);
  if (pci_fd < 0 )
  {
    perror(BC635_PATH);
    exit (1);
  }

  /*
   * map in the memory space for device registers
   */
  registers = (volatile struct stfp_registers*)mmap(0, BC635_SIZE, PROT_READ|PROT_WRITE,
                                                    MAP_SHARED, pci_fd, 0);
  if (registers == (volatile struct stfp_registers*)MAP_FAILED)
  {
    perror("mmap");
    exit (1);
  }

  /* Turn off interrupts.
   */
  registers->mask = 0;
  registers->intstat = 0xffffffff;

  /* Register the pci interrupt as a ULI source.
   */
  ULIid = (int*)ULI_register_pci(fd,       /* the pci interrupt device */
                                 intrfunc,             /* the handler function pointer */
                                 (void*)interrupt_arg, /* the argument to the handler */
                                 0,          /* the # of semaphores needed  */
                                 0,          /* the stack to use */
                                 0,          /* the stack size to use */
                                 PCI_INTA);  /* PCI interrrupt line */
  if (ULIid == 0)
  {
    perror("register uli");
    exit(1);
  }
  printf ("Registered successfully for PCI INTA - Sending interrupts\n");

  /*
   * Ask for 200 usec wait time - resolution on Origin is
   * really only ~1.5 ms instead
   */
  wait_time.tv_sec = 0;
  wait_time.tv_nsec = 200000;

  last_intr = 0;
  intr = 0;

  /* Enable the interrupt
   */
  registers->mask = 0x02000000;
  while(intr < NUM_INTS)
  {
    while (intr == last_intr)
    {
      nanosleep(&wait_time,NULL);
    }
    last_intr = intr;
  }

  printf("Number of interrupts received: %d\n", intr);
}
